# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods

"""Test cases for the runscript module"""

from __future__ import unicode_literals
from __future__ import print_function

import logging # pylint: disable=unused-import
import unittest

from helpers import bash, docker, invalid_ctype, singularity

from hpccm.primitives.runscript import runscript

class Test_runscript(unittest.TestCase):
    def setUp(self):
        """Disable logging output messages"""
        logging.disable(logging.ERROR)

    @docker
    def test_empty(self):
        """No commands specified"""
        s = runscript()
        self.assertEqual(str(s), '')

    @invalid_ctype
    def test_invalid_ctype(self):
        """Invalid container type specified"""
        s = runscript(commands=['a'])
        with self.assertRaises(RuntimeError):
            str(s)

    @docker
    def test_single_docker(self):
        """Single command specified"""
        cmd = ['z arg1 arg2']
        s = runscript(commands=cmd)
        self.assertEqual(str(s), 'ENTRYPOINT ["z", "arg1", "arg2"]')

    @singularity
    def test_single_singularity(self):
        """Single command specified"""
        cmd = ['z']
        s = runscript(commands=cmd)
        self.assertEqual(str(s), '%runscript\n    exec z "$@"')

    @docker
    def test_multiple_docker(self):
        """List of commands specified"""
        cmds = ['a', 'b', 'c']
        s = runscript(commands=cmds)
        self.assertEqual(str(s), 'ENTRYPOINT ["a"]')

    @singularity
    def test_multiple_singularity(self):
        """List of commands specified"""
        cmds = ['a arga', 'b argb', 'c']
        s = runscript(commands=cmds)
        self.assertEqual(str(s), '%runscript\n    a arga\n    b argb\n    exec c')

    @singularity
    def test_apprun_multiple_singularity(self):
        """List of commands specified"""
        cmds = ['a', 'b', 'c']
        s = runscript(commands=cmds, _app='foo')
        self.assertEqual(str(s), '%apprun foo\n    a\n    b\n    exec c')

    @docker
    def test_apprun_docker(self):
        """apprun not implemented in Docker"""
        cmds = ['a', 'b', 'c']
        s = runscript(commands=cmds, _app='foo')
        self.assertEqual(str(s), 'ENTRYPOINT ["a"]')

    @singularity
    def test_multiple_noexec_singularity(self):
        """exec option"""
        cmds = ['a', 'b', 'c']
        s = runscript(commands=cmds, _exec=False)
        self.assertEqual(str(s), '%runscript\n    a\n    b\n    c')

    @docker
    def test_merge_docker(self):
        """merge primitives"""
        r = []
        r.append(runscript(commands=['a', 'b']))
        r.append(runscript(commands=['c']))
        merged = r[0].merge(r)
        self.assertEqual(str(merged), 'ENTRYPOINT ["a"]')

    @singularity
    def test_merge_singularity(self):
        """merge primitives"""
        r = []
        r.append(runscript(commands=['a', 'b']))
        r.append(runscript(commands=['c']))
        merged = r[0].merge(r)
        self.assertEqual(str(merged), '%runscript\n    a\n    b\n    exec c')

    @bash
    def test_bash(self):
        """Single command specified"""
        cmd = ['z']
        s = runscript(commands=cmd)
        self.assertEqual(str(s), '')



# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods, bad-continuation

"""Test cases for the generic_build module"""

from __future__ import unicode_literals
from __future__ import print_function

import logging # pylint: disable=unused-import
import unittest

from helpers import centos, docker, ubuntu

from hpccm.building_blocks.generic_build import generic_build
from hpccm.toolchain import toolchain

class Test_generic_build(unittest.TestCase):
    def setUp(self):
        """Disable logging output messages"""
        logging.disable(logging.ERROR)

    @ubuntu
    @docker
    def test_defaults_ubuntu(self):
        """Default generic_build building block"""
        g = generic_build(build=['make ARCH=sm_70'],
                          install=['cp stream /usr/local/bin/cuda-stream'],
                          repository='https://github.com/bcumming/cuda-stream')
        self.assertEqual(str(g),
r'''# https://github.com/bcumming/cuda-stream
RUN mkdir -p /var/tmp && cd /var/tmp && git clone --depth=1 https://github.com/bcumming/cuda-stream cuda-stream && cd - && \
    cd /var/tmp/cuda-stream && \
    make ARCH=sm_70 && \
    cd /var/tmp/cuda-stream && \
    cp stream /usr/local/bin/cuda-stream && \
    rm -rf /var/tmp/cuda-stream''')

    @ubuntu
    @docker
    def test_no_url(self):
        """missing url"""
        with self.assertRaises(RuntimeError):
            g = generic_build()

    @ubuntu
    @docker
    def test_both_repository_and_url(self):
        """both repository and url"""
        with self.assertRaises(RuntimeError):
            g = generic_build(repository='foo', url='bar')

    @ubuntu
    @docker
    def test_invalid_package(self):
        """invalid package url"""
        with self.assertRaises(RuntimeError):
            g = generic_build(url='https://foo/bar.sh')

    @ubuntu
    @docker
    def test_prefix_recursive(self):
        """prefix and recursive option"""
        g = generic_build(build=['make ARCH=sm_70'],
                          install=['cp stream /usr/local/cuda-stream/bin/cuda-stream'],
                          prefix='/usr/local/cuda-stream/bin',
                          recursive=True,
                          repository='https://github.com/bcumming/cuda-stream')
        self.assertEqual(str(g),
r'''# https://github.com/bcumming/cuda-stream
RUN mkdir -p /var/tmp && cd /var/tmp && git clone --depth=1 --recursive https://github.com/bcumming/cuda-stream cuda-stream && cd - && \
    cd /var/tmp/cuda-stream && \
    make ARCH=sm_70 && \
    mkdir -p /usr/local/cuda-stream/bin && \
    cd /var/tmp/cuda-stream && \
    cp stream /usr/local/cuda-stream/bin/cuda-stream && \
    rm -rf /var/tmp/cuda-stream''')

    @centos
    @docker
    def test_url(self):
        """url option"""
        g = generic_build(build=['make USE_OPENMP=1'],
                          directory='OpenBLAS-0.3.6',
                          install=['make install PREFIX=/usr/local/openblas'],
                          prefix='/usr/local/openblas',
                          url='https://github.com/xianyi/OpenBLAS/archive/v0.3.6.tar.gz')
        self.assertEqual(str(g),
r'''# https://github.com/xianyi/OpenBLAS/archive/v0.3.6.tar.gz
RUN mkdir -p /var/tmp && wget -q -nc --no-check-certificate -P /var/tmp https://github.com/xianyi/OpenBLAS/archive/v0.3.6.tar.gz && \
    mkdir -p /var/tmp && tar -x -f /var/tmp/v0.3.6.tar.gz -C /var/tmp -z && \
    cd /var/tmp/OpenBLAS-0.3.6 && \
    make USE_OPENMP=1 && \
    mkdir -p /usr/local/openblas && \
    cd /var/tmp/OpenBLAS-0.3.6 && \
    make install PREFIX=/usr/local/openblas && \
    rm -rf /var/tmp/OpenBLAS-0.3.6 /var/tmp/v0.3.6.tar.gz''')

    @centos
    @docker
    def test_package(self):
        """local package"""
        g = generic_build(build=['make USE_OPENMP=1'],
                          directory='OpenBLAS-0.3.6',
                          install=['make install PREFIX=/usr/local/openblas'],
                          package='openblas/v0.3.6.tar.gz',
                          prefix='/usr/local/openblas')
        self.assertEqual(str(g),
r'''# openblas/v0.3.6.tar.gz
COPY openblas/v0.3.6.tar.gz /var/tmp/v0.3.6.tar.gz
RUN mkdir -p /var/tmp && tar -x -f /var/tmp/v0.3.6.tar.gz -C /var/tmp -z && \
    cd /var/tmp/OpenBLAS-0.3.6 && \
    make USE_OPENMP=1 && \
    mkdir -p /usr/local/openblas && \
    cd /var/tmp/OpenBLAS-0.3.6 && \
    make install PREFIX=/usr/local/openblas && \
    rm -rf /var/tmp/OpenBLAS-0.3.6 /var/tmp/v0.3.6.tar.gz''')

    @centos
    @docker
    def test_environment_ldconfig_annotate(self):
        """ldconfig and environment options"""
        g = generic_build(annotate=True,
                          base_annotation='openblas',
                          branch='v0.3.6',
                          build=['make USE_OPENMP=1'],
                          devel_environment={'CPATH': '/usr/local/openblas/include:$CPATH'},
                          install=['make install PREFIX=/usr/local/openblas'],
                          ldconfig=True,
                          prefix='/usr/local/openblas',
                          repository='https://github.com/xianyi/OpenBLAS.git',
                          runtime_environment={'CPATH': '/usr/local/openblas/include:$CPATH'})
        self.assertEqual(str(g),
r'''# https://github.com/xianyi/OpenBLAS.git
RUN mkdir -p /var/tmp && cd /var/tmp && git clone --depth=1 --branch v0.3.6 https://github.com/xianyi/OpenBLAS.git OpenBLAS && cd - && \
    cd /var/tmp/OpenBLAS && \
    make USE_OPENMP=1 && \
    mkdir -p /usr/local/openblas && \
    cd /var/tmp/OpenBLAS && \
    make install PREFIX=/usr/local/openblas && \
    echo "/usr/local/openblas/lib" >> /etc/ld.so.conf.d/hpccm.conf && ldconfig && \
    rm -rf /var/tmp/OpenBLAS
ENV CPATH=/usr/local/openblas/include:$CPATH
LABEL hpccm.openblas.branch=v0.3.6 \
    hpccm.openblas.repository=https://github.com/xianyi/OpenBLAS.git''')

        r = g.runtime()
        self.assertEqual(r,
r'''# https://github.com/xianyi/OpenBLAS.git
COPY --from=0 /usr/local/openblas /usr/local/openblas
RUN echo "/usr/local/openblas/lib" >> /etc/ld.so.conf.d/hpccm.conf && ldconfig
ENV CPATH=/usr/local/openblas/include:$CPATH
LABEL hpccm.openblas.branch=v0.3.6 \
    hpccm.openblas.repository=https://github.com/xianyi/OpenBLAS.git''')

    @ubuntu
    @docker
    def test_runtime(self):
        """Runtime"""
        g = generic_build(build=['make ARCH=sm_70'],
                          install=['cp stream /usr/local/cuda-stream/bin/cuda-stream'],
                          prefix='/usr/local/cuda-stream/bin',
                          repository='https://github.com/bcumming/cuda-stream')
        r = g.runtime()
        self.assertEqual(r,
r'''# https://github.com/bcumming/cuda-stream
COPY --from=0 /usr/local/cuda-stream/bin /usr/local/cuda-stream/bin''')

    @ubuntu
    @docker
    def test_runtime_manual(self):
        """Runtime"""
        g = generic_build(build=['make ARCH=sm_70'],
                          install=['cp stream /usr/local/cuda-stream/bin/cuda-stream'],
                          prefix='/usr/local/cuda-stream/bin',
                          repository='https://github.com/bcumming/cuda-stream',
                          runtime=['/usr/local/cuda-stream/bin'])
        r = g.runtime()
        self.assertEqual(r,
r'''# https://github.com/bcumming/cuda-stream
COPY --from=0 /usr/local/cuda-stream/bin /usr/local/cuda-stream/bin''')



"""
Basic development container image
"""

Stage0 += baseimage(image='ubuntu:18.04')

# GNU compilers
compiler = gnu()
Stage0 += compiler

# Additional development tools
Stage0 += packages(ospackages=['autoconf', 'autoconf-archive', 'automake',
                               'bzip2', 'ca-certificates', 'cmake', 'git',
                               'gzip', 'libtool', 'make', 'patch', 'xz-utils'])



# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods

"""OFED building block"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

from distutils.version import StrictVersion
import posixpath

import hpccm.config

from hpccm.building_blocks.base import bb_base
from hpccm.building_blocks.packages import packages
from hpccm.common import cpu_arch, linux_distro
from hpccm.primitives.comment import comment
from hpccm.primitives.copy import copy
from hpccm.primitives.shell import shell

class ofed(bb_base):
    """The `ofed` building block installs the OpenFabrics Enterprise
    Distribution packages that are part of the Linux distribution.

    For Ubuntu 16.04, the following packages are installed:
    `dapl2-utils`, `ibutils`, `ibverbs-utils`, `infiniband-diags`,
    `libdapl2`, `libdapl-dev`, `libibcm1`, `libibcm-dev`, `libibmad5`,
    `libibmad-dev`, `libibverbs1`, `libibverbs-dev`, `libmlx4-1`,
    `libmlx4-dev`, `libmlx5-1`, `libmlx5-dev`, `librdmacm1`,
    `librdmacm-dev`, and `rdmacm-utils`.  For Ubuntu 16.04 and aarch64
    processors, the `dapl2-utils`, `libdapl2`, `libdapl-dev`,
    `libibcm1` and `libibcm-dev` packages are not installed because
    they are not available.  For Ubuntu 16.04 and ppc64le processors,
    the `libibcm1` and `libibcm-dev` packages are not installed
    because they are not available.

    For Ubuntu 18.04, the following packages are installed:
    `dapl2-utils`, `ibutils`, `ibverbs-providers`, `ibverbs-utils`,
    `infiniband-diags`, `libdapl2`, `libdapl-dev`, `libibmad5`,
    `libibmad-dev`, `libibverbs1`, `libibverbs-dev`, `librdmacm1`,
    `librdmacm-dev`, and `rdmacm-utils`.

    For RHEL-based 7.x distributions, the following packages are
    installed: `dapl`, `dapl-devel`, `ibutils`, `libibcm`, `libibmad`,
    `libibmad-devel`, `libmlx5`, `libibumad`, `libibverbs`,
    `libibverbs-utils`, `librdmacm`, `rdma-core`, and
    `rdma-core-devel`.

    For RHEL-based 8.x distributions, the following packages are
    installed: `libibmad`, `libibmad-devel`, `libmlx5`, `libibumad`,
    `libibverbs`, `libibverbs-utils`, `librdmacm`, `rdma-core`, and
    `rdma-core-devel`.

    # Parameters

    prefix: The top level install location. Install of installing the
    packages via the package manager, they will be extracted to this
    location. This option is useful if multiple versions of OFED need
    to be installed. The environment must be manually configured to
    recognize the OFED location, e.g., in the container entry
    point. The default value is empty, i.e., install via the package
    manager to the standard system locations.

    # Examples

    ```python
    ofed()
    ```

    """

    def __init__(self, **kwargs):
        """Initialize building block"""

        super(ofed, self).__init__(**kwargs)

        self.__deppackages = []  # Filled in by __distro()
        self.__extra_opts = []   # Filled in by __distro()
        self.__ospackages = []   # Filled in by __distro()
        self.__powertools = False # enable the CentOS PowerTools repo
        self.__prefix = kwargs.get('prefix', None)
        self.__symlink = kwargs.get('symlink', False)
        self.__wd = kwargs.get('wd', hpccm.config.g_wd) # working directory

        # Set the Linux distribution specific parameters
        self.__distro()

        # Fill in container instructions
        self.__instructions()

    def __distro(self):
        """Based on the Linux distribution, set values accordingly.  A user
           specified value overrides any defaults."""

        if hpccm.config.g_linux_distro == linux_distro.UBUNTU:
            self.__deppackages = ['libnl-3-200', 'libnl-route-3-200',
                                 'libnuma1']

            if hpccm.config.g_linux_version >= StrictVersion('18.0'):
                # Give priority to packages from the Ubuntu repositories over
                # vendor repositories
                if hpccm.config.g_linux_version >= StrictVersion('20.0'):
                    self.__extra_opts = ['-t focal']
                else:
                    self.__extra_opts = ['-t bionic']

                self.__ospackages= ['dapl2-utils', 'ibutils',
                                    'ibverbs-providers', 'ibverbs-utils',
                                    'infiniband-diags',
                                    'libdapl2', 'libdapl-dev',
                                    'libibmad5', 'libibmad-dev',
                                    'libibverbs1', 'libibverbs-dev',
                                    'librdmacm1', 'librdmacm-dev',
                                    'rdmacm-utils']
            else:
                # Give priority to packages from the Ubuntu repositories over
                # vendor repositories
                self.__extra_opts = ['-t xenial']

                self.__ospackages = ['dapl2-utils', 'ibutils', 'ibverbs-utils',
                                     'infiniband-diags',
                                     'libdapl2', 'libdapl-dev',
                                     'libibcm1', 'libibcm-dev',
                                     'libibmad5', 'libibmad-dev',
                                     'libibverbs1', 'libibverbs-dev',
                                     'libmlx4-1', 'libmlx4-dev',
                                     'libmlx5-1', 'libmlx5-dev',
                                     'librdmacm1', 'librdmacm-dev',
                                     'rdmacm-utils']

                if hpccm.config.g_cpu_arch == cpu_arch.AARCH64:
                    # Ubuntu 16.04 for ARM is missing these packages
                    for missing in ['dapl2-utils', 'libdapl2', 'libdapl-dev',
                                    'libibcm1', 'libibcm-dev']:
                        if missing in self.__ospackages:
                            self.__ospackages.remove(missing)
                elif hpccm.config.g_cpu_arch == cpu_arch.PPC64LE:
                    # Ubuntu 16.04 for Power is missing these packages
                    for missing in ['libibcm1', 'libibcm-dev']:
                        if missing in self.__ospackages:
                            self.__ospackages.remove(missing)
        elif hpccm.config.g_linux_distro == linux_distro.CENTOS:
            self.__extra_opts = [r'--disablerepo=mlnx\*']

            if hpccm.config.g_linux_version >= StrictVersion('8.0'):
                self.__deppackages = ['libnl3', 'numactl-libs']
                self.__ospackages = ['libibmad', 'libibmad-devel',
                                     'libibumad', 'libibverbs',
                                     'libibverbs-utils', 'libmlx5',
                                     'librdmacm',
                                     'rdma-core', 'rdma-core-devel']
                self.__powertools = True
            else:
                self.__deppackages = ['libnl', 'libnl3', 'numactl-libs']
                self.__ospackages = ['dapl', 'dapl-devel', 'ibutils',
                                     'libibcm', 'libibmad', 'libibmad-devel',
                                     'libmlx5', 'libibumad', 'libibverbs',
                                     'libibverbs-utils', 'librdmacm',
                                     'rdma-core', 'rdma-core-devel']
        else: # pragma: no cover
            raise RuntimeError('Unknown Linux distribution')

    def __instructions(self):
        """Fill in container instructions"""
        self += comment('OFED')
        if self.__prefix:
            commands = []

            # Extract to a prefix - not a "real" package manager install
            self += packages(ospackages=self.__deppackages)
            self += packages(download=True, extra_opts=self.__extra_opts,
                             extract=self.__prefix,
                             ospackages=self.__ospackages,
                             powertools=self.__powertools)

            # library symlinks
            if self.__symlink:
                self.__deppackages.append('findutils')

                commands.append('mkdir -p {0} && cd {0}'.format(
                    posixpath.join(self.__prefix, 'lib')))
                # Prune the symlink directory itself and any debug
                # libraries
                commands.append('find .. -path ../lib -prune -o -name "*valgrind*" -prune -o -name "lib*.so*" -exec ln -s {} \;')
                commands.append('cd {0} && ln -s usr/bin bin && ln -s usr/include include'.format(
                    self.__prefix))

            # Suppress warnings from libibverbs
            commands.append('mkdir -p /etc/libibverbs.d')

            self += shell(commands=commands)
        else:
            # Install packages using package manager
            self += packages(extra_opts=self.__extra_opts,
                             ospackages=self.__ospackages,
                             powertools=self.__powertools)

    def runtime(self, _from='0'):
        """Generate the set of instructions to install the runtime specific
        components from a build in a previous stage.

        # Examples

        ```python
        o = ofed(...)
        Stage0 += o
        Stage1 += o.runtime()
        ```
        """
        if self.__prefix:
            self.rt += comment('OFED')

            if self.__deppackages:
                self.rt += packages(ospackages=self.__deppackages)

            # Suppress warnings from libibverbs
            self.rt += shell(commands=['mkdir -p /etc/libibverbs.d'])

            self.rt += copy(_from=_from, dest=self.__prefix, src=self.__prefix)
            return str(self.rt)
        else:
            return str(self)
